/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mips.h"
#include "pic32.h"

#define LINE_LENGTH	160	/* pixels */

#define HFREQ_LIMIT	1254	/* 40MHz cycles */
#define HSYNC_START	800	/* 40MHz cycles */
#define HSYNC_LIMIT	96	/* 40MHz cycles */
#define HSYNC_END	(HSYNC_START + HSYNC_LIMIT)

#define VISIBLE_START	15	/* horizontal lines, back porch end */
#define VFP_START	527	/* horizontal lines, front porch start */
#define VSYNC_START	529	/* horizontal lines, front porch end */
#define VSYNC_END	531	/* horizontal lines, back porch start */

#define SCREEN_SIZE	(40 * 1024)

/* Disable JTAG functionality on pins. */

.section .devcfg0, "a"
.word 0xfffffffb	/* DEVCFG0<2> = JTAGEN = 0 */

/*
Set the oscillator to be the FRC oscillator with PLL, with peripheral clock
divided by 1, and FRCDIV+PLL selected.

The system clock and peripheral clock are therefore the same.

The watchdog timer (FWDTEN) is also disabled.

The secondary oscillator pin (FSOSCEN) is disabled to avoid pin conflicts with
RPB4.
*/

.section .devcfg1, "a"
.word 0xff7fcfd9	/* DEVCFG1<23> = FWDTEN = 0; DEVCFG1<13:12> = FPBDIV<2:0> = 0;
			DEVCFG1<5> = FSOSCEN = 0; DEVCFG1<2:0> = FNOSC<2:0> = 001 */

/*
Set the FRC oscillator PLL function with an input division of 4, an output
division of 2, a multiplication of 20, yielding a multiplication of 2.5.

The FRC is apparently at 16MHz and this produces a system clock of 40MHz.
*/

.section .devcfg2, "a"
.word 0xfff9ffdb	/* DEVCFG2<18:16> = FPLLODIV<2:0> = 001;
			DEVCFG2<6:4> = FPLLMUL<2:0> = 101;
			DEVCFG2<2:0> = FPLLIDIV<2:0> = 011 */

.text
.globl _start

_start:
	/*
	Configure RAM.
	See: http://microchipdeveloper.com/32bit:mx-arch-exceptions-processor-initialization
	*/

	la $v0, BMXCON
	lw $v1, 0($v0)
	li $t8, ~(1 << 6)	/* BMXCON<6> = BMXWSDRM = 0 */
	and $v1, $v1, $t8
	li $t8, ~0b111		/* BMXCON<2:0> = BMXARB<2:0> = 0 */
	ori $t8, $t8, 0b010	/* BMXCON<2:0> = BMXARB<2:0> = 2 */
	and $v1, $v1, $t8
	sw $v1, 0($v0)

	/* Enable caching. */

	li $v0, CONFIG_CM_CACHABLE_NONCOHERENT
	mtc0 $v0, CP0_CONFIG
	nop

	/* Get the RAM size. */

	la $v0, BMXDRMSZ
	lw $v0, 0($v0)

	/* Initialise the stack pointer. */

	li $v1, KSEG0_BASE
	addu $sp, $v0, $v1	/* sp = KSEG0_BASE + RAM size */

	/* Initialise framebuffer. */

	jal init_framebuffer
	nop

	/* Initialise the globals pointer. */

	lui $gp, %hi(_GLOBAL_OFFSET_TABLE_)
	ori $gp, $gp, %lo(_GLOBAL_OFFSET_TABLE_)

	/* Set pins for output. */

	jal init_pins
	nop

	jal init_oc_pins
	nop

	/* Initialise the status register. */

	jal init_interrupts
	nop

	/* Initialise timer. */

	jal init_timer2
	nop

	/* Initialise PMP. */

	jal init_pmp
	nop

	/* Initialise DMA. */

	jal init_dma
	nop

	/* Initialise OC3 and OC2. */

	jal init_oc
	nop

	/* Enable interrupts and loop. */

	jal enable_interrupts
	nop

	jal handle_error_level
	nop

	/* Set initial sync conditions. */

	la $t0, PORTB
	li $t1, (1 << 5)	/* PORTB<5> = RB5 */
	sw $t1, SET($t0)

	la $t0, PORTB
	li $t1, (1 << 10)	/* PORTB<10> = RB10 */
	sw $t1, SET($t0)

	/* Main program. */

	li $a1, 5000000		/* counter = 5000000 */

	/* Initialise the display state. */

	li $s0, 0		/* line counter */
	la $s1, vbp_active	/* current event */
	move $s2, $zero		/* line address */

	/* Monitoring loop. */
loop:
	addiu $a1, $a1, -1	/* counter -= 1 */
	bnez $a1, loop		/* until counter == 0 */
	nop

	li $a1, 5000000		/* counter = 5000000 */

	la $t0, PORTB
	li $t1, (1 << 11)	/* PORTB<11> = RB11 */
	sw $t1, INV($t0)

_next:
	j loop
	nop



init_pins:
	/* DEVCFG0<2> needs setting to 0 before the program is run. */

	la $v0, CFGCON
	li $v1, (1 << 3)	/* CFGCON<3> = JTAGEN = 0 */
	sw $v1, CLR($v0)

init_outputs:
	/* Remove analogue features from pins. */

	la $v0, ANSELA
	sw $zero, 0($v0)	/* ANSELA = 0 */
	la $v0, ANSELB
	sw $zero, 0($v0)	/* ANSELB = 0 */

	la $v0, TRISA
	sw $zero, 0($v0)
	la $v0, TRISB
	sw $zero, 0($v0)

	la $v0, PORTA
	sw $zero, 0($v0)
	la $v0, PORTB
	sw $zero, 0($v0)

	jr $ra
	nop



/* Initialisation routines. */

init_timer2:

	/* Initialise Timer2 interrupt. */

	la $v0, T2CON
	sw $zero, 0($v0)	/* T2CON = 0 */
	nop

	la $v0, TMR2
	sw $zero, 0($v0)	/* TMR2 = 0 */

	la $v0, PR2
	li $v1, HFREQ_LIMIT
	sw $v1, 0($v0)		/* PR2 = HFREQ_LIMIT */

	/* Initialise Timer2 interrupt. */

	la $v0, IFS0
	li $v1, (1 << 9)
	sw $v1, CLR($v0)	/* T2IF = 0 */

	la $v0, IPC2
	li $v1, 0b11111
	sw $v1, CLR($v0)	/* T2IP, T2IS = 0 */

	la $v0, IPC2
	li $v1, 0b11111
	sw $v1, SET($v0)	/* T2IP = 7; T2IS = 3 */

	la $v0, IEC0
	li $v1, (1 << 9)
	sw $v1, SET($v0)	/* T2IE = 1 */

	/* Start timer. */

	la $v0, T2CON
	li $v1, (1 << 15)
	sw $v1, SET($v0)	/* ON = 1 */

	jr $ra
	nop



/*
Output compare initialisation.

Timer2 will be used to trigger two events using OC3: one initiating the hsync
pulse, and one terminating the pulse. The pulse should appear after the line
data has been transferred using DMA, but this is achieved by just choosing
suitable start and end values.

Using OC2, Timer 2 triggers a level shifting event and OC2 is reconfigured to
reverse the level at a later point.
*/

init_oc:
	/* Disable OC3 interrupts. */

	la $v0, IEC0
	li $v1, (1 << 17)	/* IEC0<17> = OC3IE = 0 */
	sw $v1, CLR($v0)

	la $v0, IFS0
	li $v1, (1 << 17)	/* IFS0<17> = OC3IF = 0 */
	sw $v1, CLR($v0)

	/* Initialise OC3. */

	la $v0, OC3CON
	li $v1, 0b101		/* OC3CON<2:0> = OCM<2:0> = 101 (dual compare, continuous pulse) */
	sw $v1, 0($v0)

	/* Pulse start and end. */

	la $v0, OC3R
	li $v1, HSYNC_END	/* HSYNC_START for positive polarity */
	sw $v1, 0($v0)

	la $v0, OC3RS
	li $v1, HSYNC_START	/* HSYNC_END for positive polarity */
	sw $v1, 0($v0)

	/* OC3 is enabled. */

	la $v0, OC3CON
	li $v1, (1 << 15)
	sw $v1, SET($v0)

	/* Disable OC2 interrupts. */

	la $v0, IEC0
	li $v1, (1 << 12)	/* IEC0<12> = OC2IE = 0 */
	sw $v1, CLR($v0)

	la $v0, IFS0
	li $v1, (1 << 12)	/* IFS0<12> = OC2IF = 0 */
	sw $v1, CLR($v0)

	/* Initialise OC2. */

	la $v0, OC2CON
	li $v1, 0b010		/* OC2CON<2:0> = OCM<2:0> = 010 (single compare, output driven low) */
	sw $v1, 0($v0)

	/* Set pulse position. */

	la $v0, OC2R
	sw $zero, 0($v0)

	/* Enable OC2 later. */

	jr $ra
	nop


init_oc_pins:
	/* Unlock the configuration register bits. */

	la $v0, SYSKEY
	sw $zero, 0($v0)
	li $v1, 0xAA996655
	sw $v1, 0($v0)
	li $v1, 0x556699AA
	sw $v1, 0($v0)

	la $v0, CFGCON
	lw $t8, 0($v0)
	li $v1, (1 << 13)	/* IOLOCK = 0 */
	sw $v1, CLR($v0)

	/* Map OC3 to RPB10. */

	la $v0, RPB10R
	li $v1, 0b0101		/* RPB10R<3:0> = 0101 (OC3) */
	sw $v1, 0($v0)

	/* Map OC2 to RPB5. */

	la $v0, RPB5R
	li $v1, 0b0101		/* RPB5R<3:0> = 0101 (OC2) */
	sw $v1, 0($v0)

	la $v0, CFGCON
	sw $t8, 0($v0)

	/* Lock the oscillator control register again. */

	la $v0, SYSKEY
	li $v1, 0x33333333
	sw $v1, 0($v0)

	jr $ra
	nop



/* Parallel Master Port initialisation. */

init_pmp:
	/* Disable PMP interrupts. */

	la $v0, IEC1
	li $v1, (1 << 16)	/* IEC1<16> = PMPIE = 0 */
	sw $v1, CLR($v0)

	/* Initialise PMP.

	PMCON<12:11> = ADDRMUX<1:0> = 0; demultiplexed address and data
	PMCON<9> = PTWREN<0> = 0; no write pin
	PMCON<8> = PTRDEN<0> = 0; no read pin
	PMCON<7:6> = CSF<1:0> = 0; no chip select pins
	*/

	la $v0, PMCON
	sw $zero, 0($v0)

	/*
	PMMODE<14:13> = IRQM<1:0> = 1; interrupt after every read/write
	PMMODE<12:11> = INCM<1:0> = 0; no increment on every read/write
	PMMODE<10> = MODE16<0> = 0; 8-bit transfers
	PMMODE<9:8> = MODE<1:0> = 10; master mode 2
	PMMODE<5:2> = WAITM<3:0> = 00; single cycle read/write, no chip select wait cycles
	*/

	la $v0, PMMODE
	li $v1, 0x2200
	sw $v1, 0($v0)

	/* Free non-essential pins for general I/O. */

	la $v0, PMAEN
	sw $zero, 0($v0)

	la $v0, PMADDR
	sw $zero, 0($v0)

	la $v0, IFS1
	li $v1, (3 << 16)	/* IFS1<17:16> = PMPEIF, PMPIF = 0 */
	sw $v1, CLR($v0)

	/* Start PMP mode. */

	la $v0, PMCON
	li $v1, (1 << 15)	/* PMCON<15> = ON = 1 */
	sw $v1, SET($v0)

	jr $ra
	nop



/*
Direct Memory Access initialisation.

Write 160 pixels to the PMP for the line data. This is initiated by a timer
interrupt. Upon completion of the transfer, a DMA interrupt initiates the
address update routine, changing the source address of the DMA channel.
*/

init_dma:
	/* Disable DMA interrupts. */

	la $v0, IEC1
	li $v1, (1 << 28)	/* IEC1<28> = DMA0IE = 0 */
	sw $v1, CLR($v0)

	/* Clear DMA interrupt flags. */

	la $v0, IFS1
	li $v1, (1 << 28)	/* IFS1<28> = DMA0IF = 0 */
	sw $v1, CLR($v0)

	/* Enable DMA. */

	la $v0, DMACON
	li $v1, (1 << 15)
	sw $v1, SET($v0)

	/*
	Initialise a line channel.
	The line channel will be channel 0 (x = 0).

	Specify a priority of 3:
	DCHxCON<1:0> = CHPRI<1:0> = 3

	Auto-enable the channels:
	DCHxCON<4> = CHAEN = 1
	*/

	la $v0, DCH0CON
	li $v1, 0b10011
	sw $v1, 0($v0)

	/*
	Initiate channel transfers when the initiating interrupt condition
	occurs:
	DCHxECON<15:8> = CHSIRQ<7:0> = timer 2 interrupt
	DCHxECON<4> = SIRQEN = 1
	*/

	la $v0, DCH0ECON
	li $v1, (9 << 8) | (1 << 4)
	sw $v1, 0($v0)

	/*
	The line channel has a cell size as the number bytes in a line:
	DCHxCSIZ<15:0> = CHCSIZ<15:0> = LINE_LENGTH
	*/

	la $v0, DCH0CSIZ
	li $v1, LINE_LENGTH
	sw $v1, 0($v0)

	/*
	The source has a size identical to the cell size:
	DCHxSSIZ<15:0> = CHSSIZ<15:0> = LINE_LENGTH
	*/

	la $v0, DCH0SSIZ
	li $v1, LINE_LENGTH
	sw $v1, 0($v0)

	/*
	The source address is the physical address of the line data:
	DCHxSSA = physical(line data address)
	*/

	la $v0, DCH0SSA
	sw $zero, 0($v0)

	/*
	The destination has a size of 1 byte:
	DCHxDSIZ<15:0> = CHDSIZ<15:0> = 1
	*/

	la $v0, DCH0DSIZ
	li $v1, 1
	sw $v1, 0($v0)

	/*
	The destination address is the physical address of PMDIN:
	DCHxDSA = physical(PMDIN)
	*/

	la $v0, DCH0DSA
	li $v1, PMDIN
	li $t8, KSEG1_BASE
	subu $v1, $v1, $t8
	sw $v1, 0($v0)

	/*
	Use the block transfer completion interrupt to indicate when the source
	address can be updated.
	*/

	la $v0, DCH0INT
	li $v1, (1 << 19)	/* CHBCIE = 1 */
	sw $v1, 0($v0)

	/* Enable interrupt for address updating. */

	la $v0, IPC10
	li $v1, 0b11111		/* DMA0IP, DMA0IS = 0 */
	sw $v1, CLR($v0)

	la $v0, IPC10
	li $v1, 0b11111		/* DMA0IP = 7, DMA0IS = 3 */
	sw $v1, SET($v0)

	la $v0, IEC1
	li $v1, (1 << 28)	/* IEC1<28> = DMA0IE = 1 */
	sw $v1, SET($v0)

	/* Enable channel. */

	la $v0, DCH0CON
	li $v1, 0b10000000
	sw $v1, SET($v0)

	jr $ra
	nop



/* Framebuffer initialisation. */

init_framebuffer:
	li $v0, KSEG0_BASE
	li $t8, SCREEN_SIZE
	li $v1, 0xff031ce0

_init_fb_loop:
	sw $v1, 0($v0)
	addiu $v0, $v0, 4
	addiu $t8, $t8, -4
	bnez $t8, _init_fb_loop
	nop

	jr $ra
	nop



/* Utilities. */

handle_error_level:
	mfc0 $t3, CP0_STATUS
	li $t4, ~(STATUS_ERL | STATUS_EXL)
	and $t3, $t3, $t4
	mtc0 $t3, CP0_STATUS
	jr $ra
	nop

enable_interrupts:
	mfc0 $t3, CP0_STATUS
	li $t4, ~STATUS_IRQ	/* Clear interrupt priority bits. */
	and $t3, $t3, $t4
	li $t4, ~STATUS_BEV	/* CP0_STATUS &= ~STATUS_BEV (use non-bootloader vectors) */
	and $t3, $t3, $t4
	ori $t3, $t3, STATUS_IE
	mtc0 $t3, CP0_STATUS
	jr $ra
	nop

init_interrupts:
	mfc0 $t3, CP0_DEBUG
	li $t4, ~DEBUG_DM
	and $t3, $t3, $t4
	mtc0 $t3, CP0_DEBUG

	mfc0 $t3, CP0_STATUS
	li $t4, STATUS_BEV	/* BEV = 1 or EBASE cannot be set */
	or $t3, $t3, $t4
	mtc0 $t3, CP0_STATUS

	la $t3, exception_handler
	mtc0 $t3, CP0_EBASE	/* EBASE = exception_handler */

	li $t3, 0x20		/* Must be non-zero or the CPU gets upset */
	mtc0 $t3, CP0_INTCTL

	li $t3, CAUSE_IV	/* IV = 1 (use EBASE+0x200 for interrupts) */
	mtc0 $t3, CP0_CAUSE

	jr $ra
	nop



/* Exception servicing. */

.section .flash, "a"

exception_handler:
	li $t8, 2500000
exc_loop:
	addiu $t8, $t8, -1
	bnez $t8, exc_loop
	nop
	la $v0, PORTB
	li $v1, (1 << 11)	/* PORTB<11> = RB11 */
	sw $v1, INV($v0)
	j exception_handler
	nop



/* Interrupt servicing. */

.org 0x200

interrupt_handler:

	/* Check for a timer interrupt condition. */

	la $v0, IFS0
	lw $v1, 0($v0)
	andi $v1, $v1, (1 << 9)	/* T2IF */
	beqz $v1, irq_dma
	nop

	/* Increment the line counter. */

	addiu $s0, $s0, 1

	/* Jump to the event handler. */

	jalr $s1
	nop

irq_clear_timer:

	/* Clear the timer interrupt condition. */

	la $v0, IFS0
	li $v1, (1 << 9)	/* IFS0<9> = T2IF = 0 */
	sw $v1, CLR($v0)

irq_dma:

	/* Check for a DMA interrupt condition. */

	la $v0, IFS1
	lw $v1, 0($v0)
	li $t8, (1 << 28)	/* DMA0IF */
	and $v1, $v1, $t8
	beqz $v1, irq_exit
	nop

	/* Test the block transfer completion interrupt flag. */

	la $v0, DCH0INT
	lw $v1, 0($v0)
	andi $v1, $v1, (1 << 3)	/* CHBCIF */
	beqz $v1, irq_clear_dma
	nop

	/* Clear the block transfer completion interrupt flag. */

	li $v1, (1 << 3)	/* CHBCIF = 0 */
	sw $v1, CLR($v0)

	/*
	Update the line data address if the line counter (referring to the
	next line) is even.
	*/

	andi $t8, $s0, 1
	bnez $t8, irq_clear_dma
	nop

	/* Reference the next line and update the DMA source address. */

	addiu $s2, $s2, LINE_LENGTH

	/* Test for wraparound. */

	li $t8, SCREEN_SIZE
	sltu $t8, $s2, $t8
	bnez $t8, irq_dma_update
	nop

	/* Reset the source address. */

	move $s2, $zero

irq_dma_update:

	la $v0, DCH0SSA
	sw $s2, 0($v0)

irq_clear_dma:

	/* Clear the DMA interrupt condition. */

	la $v0, IFS1
	li $v1, (1 << 28)	/* IFS1<28> = DMA0IF = 0 */
	sw $v1, CLR($v0)

irq_exit:
	eret
	nop



/* Event routines. */

/* The vertical back porch. */

vbp_active:
	/* Test for visible region. */

	sltiu $v0, $s0, VISIBLE_START
	bnez $v0, _vbp_active_ret
	nop

	/* Start the visible region. */

	la $s1, visible_active

_vbp_active_ret:
	jr $ra
	nop



/* The visible region. */

visible_active:
	/* Test for front porch. */

	sltiu $v0, $s0, VFP_START
	bnez $v0, _visible_active_ret
	nop

	/* Start the front porch region. */

	la $s1, vfp_active

_visible_active_ret:
	jr $ra
	nop



/* Within the vertical front porch. */

vfp_active:
	/* Test for vsync. */

	sltiu $v0, $s0, VSYNC_START
	bnez $v0, _vfp_active_ret
	nop

	/* Start the vsync. */

	la $s1, vsync_active

	/* Bring vsync low when the next line starts. */

	la $v0, OC2CON
	li $v1, 0b010 | (1 << 15)	/* OC2CON<2:0> = OCM<2:0> = 010 (single compare, output driven low) */
	sw $v1, 0($v0)

_vfp_active_ret:
	jr $ra
	nop



/* The vsync period. */

vsync_active:
	/* Test for front porch. */

	sltiu $v0, $s0, VSYNC_END
	bnez $v0, _vsync_active_ret
	nop

	/* Start the back porch. */

	move $s0, $zero
	la $s1, vbp_active

	/* Bring vsync high when the next line starts. */

	la $v0, OC2CON
	li $v1, 0b001 | (1 << 15)	/* OC2CON<2:0> = OCM<2:0> = 001 (single compare, output driven high) */
	sw $v1, 0($v0)

_vsync_active_ret:
	jr $ra
	nop
